<p>A password reset has been requested for {{ email }}. If you didn't make this request, you can ignore this email.</p>

<p>To reset your password, please click the link below:</p>

<p><a href="{{ url }}">{{ url }}</a></p>
